//-----------------------------------------------------------------------------
// QUEUE.H
// By Dean Harding
// 
// Copyright (c) 2000/2001 Dean Harding
//-----------------------------------------------------------------------------
// I needed a fast queue system, where it's very fast *adding* to the *end* of
// the queue, and *removing* from the beginning (i.e. FIFO).  I don't know of
// any STL container that have those properties, so I made my own.
//
// It's just a doubly-linked list :)
//
// Also, this class is thread safe (a thread can add to the queue while another
// is getting stuff from the top)  And finally, you can block on a GetNext()
// call until data is added.  Note that this only works when there is one
// *and only one* thread using GetNext().  That's OK - that's exactly the
// functionality we want!
//-----------------------------------------------------------------------------
#pragma once

//-----------------------------------------------------------------------------
template< class T >
class CQueue
{
protected:
	struct SQueueData
	{
		T *pData;

		SQueueData *pNext;
		SQueueData *pPrev;
	};

	SQueueData *m_pTop;
	SQueueData *m_pBottom;

	// We just lock the whole queue when we want to add/remove to/from it,
	// because that's nice and easy, and besides, the things we do on the queue
	//should be simple enough that it won't really matter.
	HANDLE m_hQueueLock;

	// When we add an item queue, this even it fired to tell anything waiting
	// in GetNext().
	HANDLE m_hAddEvent;

	// This is used by WaitForEmpty().  When GetNext() removes the last item
	// from the queue, this is signalled.
	HANDLE m_hEmptyEvent;
public:
	inline CQueue();
	inline ~CQueue();

	// Get the first item off the top of the queue.  If the queue is empty
	// and bBlock is false, NULL is returned.  If the queue is empty and bBlock
	// is true, then the function blocks until the queue is added to.
	inline T* GetNext( bool bBlock = false );

	// Add an item to the queue.  If there's a blocking call to GetNext() then
	// that call will wake up and return this item.
	inline void AddItem( T* pItem );

	// returns true if the queue is empty.
	inline bool IsEmpty();

	// blocks until the queue is empty
	inline void WaitForEmpty();
};

//-----------------------------------------------------------------------------
#include "queue.inl"

//-----------------------------------------------------------------------------
// End of QUEUE.H
//-----------------------------------------------------------------------------
